/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.example.order.manager.service;

import java.text.DateFormat;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bluestemsoftware.open.eoa.example.order.manager.service.dao.CustomerOrderDAO;
import org.bluestemsoftware.open.eoa.example.order.manager.service.error.SystemError;
import org.bluestemsoftware.open.eoa.example.order.manager.service.fault.OrderNotFoundFault;
import org.bluestemsoftware.open.eoa.example.order.manager.service.fault.OutOfStockFault;
import org.bluestemsoftware.open.eoa.example.order.manager.service.vo.CustomerOrder;
import org.bluestemsoftware.open.eoa.example.order.manager.service.vo.Item;
import org.bluestemsoftware.open.eoa.example.order.manager.service.vo.OutstandingRequest;
import org.bluestemsoftware.open.eoa.example.order.manager.service.vo.CustomerOrder.Status;
import org.bluestemsoftware.open.eoa.example.order.manager.util.Constants;
import org.bluestemsoftware.open.eoa.example.order.manager.util.DOMUtils;
import org.bluestemsoftware.open.eoa.example.order.manager.vo.AvailabilityResponse;
import org.bluestemsoftware.specification.eoa.application.spring.MyApplication;
import org.bluestemsoftware.specification.eoa.application.spring.MyApplicationContext;
import org.bluestemsoftware.specification.eoa.application.spring.MyOperation;
import org.bluestemsoftware.specification.eoa.application.spring.MyRole;
import org.bluestemsoftware.specification.eoa.application.spring.PartnerApplication;
import org.bluestemsoftware.specification.eoa.application.spring.PartnerOperation;
import org.bluestemsoftware.specification.eoa.application.spring.PartnerRole;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ApplicationFault;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.FaultContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Encapsulates logic related to order fulfillment within context of order manager application.
 * Note that this service exposes a web-service interface.
 */
public class OrderProcessorImpl implements OrderProcessor {

    private static final Log log = LogFactory.getLog(OrderProcessor.class);
    private static final String ROLE_NAME = "http://mycompany.com/om/order/processor";

    private CustomerOrderDAO customerOrderDAO;
    private CatalogProcessor catalogProcessor;
    private DocumentBuilder documentBuilder;
    private MyApplication myApplication;
    private MyRole myRole;

    /**
     * Selects partner application, i.e. an executable instance of WarehouseManger application,
     * based upon closest proximity to customer zipcode and places order.
     * <p>
     * Note that this method utilizes the preferred method signature which employs a
     * non-blocking invocation and which provides access to 'middleware' api, e.g. to
     * {@link ActionContext} and {@link MessageRT}.
     * 
     * 
     * @param myOperation
     * @param requestContext
     */
    public void placeOrder(MyOperation myOperation, ActionContext requestContext) {

        // using message fragments (which map a schema type to a
        // predefined query string) we extract the zipcode and
        // items element from request

        Message request = requestContext.getMessage();
        String zipCode = request.evaluateString(Constants.ZIP_CODE);
        Element itemsEl = (Element)request.evaluateNode(Constants.ITEMS);
        List<Element> itemsList = DOMUtils.getChildElements(itemsEl);

        // using customer zip, select warehouse within closest proximity to
        // customer location and retrieve proxy which models the warehouse
        // management system

        WarehouseLocation warehouseManager = WarehouseLocation.getInstance(zipCode);

        // check to see if items are available to promise. if not, an
        // out of stock exception is thrown in which case we return
        // an out of stock fault to caller

        AvailabilityResponse response = null;
        try {
            response = warehouseManager.checkAvailability(zipCode, itemsList);
        } catch (OutOfStockFault of) {
            FaultContext fResponse = myOperation.createFaultAction(Constants.OUT_OF_STOCK_FAULT);
            Message faultMessage = fResponse.getMessage();
            Element temp = (Element)faultMessage.evaluateNode(Constants.ITEMS);
            for (Element item : DOMUtils.getChildElements(of.getPayload())) {
                temp.appendChild(faultMessage.adoptNode(item));
            }
            myOperation.sendResponse(fResponse);
            return;
        }

        // items are in stock and have been promised to customer. using
        // customer order number generated by warehouse management
        // system, create and persist order on our system

        Set<Item> items = new HashSet<Item>();
        for (Element temp : itemsList) {
            Item catalogItem = catalogProcessor.getItem(temp.getAttribute("id"));
            Item orderItem = new Item(catalogItem.getId(), catalogItem.getDescription());
            orderItem.setQuantity(Integer.parseInt(temp.getAttribute("quantity")));
            orderItem.setPrice(catalogItem.getPrice() * orderItem.getQuantity());
            items.add(orderItem);
        }

        String orderNumber = response.getOrderNumber();
        double sc = response.getShippingCost();
        CustomerOrder customerOrder = new CustomerOrder(orderNumber, zipCode, items, sc);
        customerOrderDAO.insertCustomerOrder(customerOrder);

        // now we schedule the order with warehouse management system.
        // note that this is a potentially long running exchange, so
        // we 'register' the message exchange instance on customer
        // order as an outstanding request which is unmarshalled
        // when partner response is received and used to complete
        // the exchange within placeOrderResponse

        customerOrder.addOutstandingRequest(new OutstandingRequest(myOperation));
        customerOrderDAO.updateCustomerOrder(customerOrder);
        warehouseManager.scheduleOrder(orderNumber);

    }

    /**
     * Invoked by AbstractWarehouse which processes asynchronous replies sent in response to
     * schedule order requests. Completes outstanding placeOrder requests.
     * 
     * @param customerOrder
     * @param estimatedShipDate
     */
    public void placeOrderResponse(CustomerOrder customerOrder, Date estimatedShipDate) {

        // now we process the outstanding process order request
        // by returning process order response to client

        String mexID = MyOperation.toID(myRole, Constants.PLACE_ORDER_OPERATION);
        OutstandingRequest or = customerOrder.getOutstandingRequest(mexID);
        MyOperation myOperation = or.getMessageExchange();

        // create a runtime response message instance. the def for
        // which is provided within the example-business-docs
        // deployment descriptor, i.e. we don't have to build it
        // up manually

        ActionContext placeOrderResponse = myOperation.createResponseAction();
        Element payload = placeOrderResponse.getMessage().getContent();
        payload.setAttribute("orderNumber", customerOrder.getOrderNumber());
        String date = DateFormat.getDateInstance().format(customerOrder.getEstShipDate());
        payload.setAttribute("estimatedShipDate", date);
        double totalPrice = 0;
        for (Item orderItem : customerOrder.getItems()) {
            totalPrice = totalPrice + orderItem.getPrice();
        }
        double shipCost = customerOrder.getShipCost();
        totalPrice = totalPrice + shipCost;
        payload.setAttribute("shippingCost", Double.toString(shipCost));
        payload.setAttribute("totalPrice", Double.toString(totalPrice));
        myOperation.sendResponse(placeOrderResponse);
        or.setResponseSent(true);
        customerOrderDAO.updateCustomerOrder(customerOrder);

    }

    /**
     * An example method signature which does not rely upon classes defined within 'middleware'
     * api, i.e. classes defined within the EOA specification api.
     * <p>
     * Note method signature employs a blocking invocation. See
     * {@link #placeOrder(MyOperation, ActionContext)} for signature which provides access to
     * runtime request/response message defs, predefined message fragments and which
     * facilitates a non-blocking invocation for long running exchanges.
     * 
     * @param request
     * @return order status response
     * @throws OrderNotFoundFault
     *         if order not found
     */
    public Element getOrderStatus(Element request) throws OrderNotFoundFault {

        // retrieve order number for which status is being requested.
        // order was created within place order method

        QName childName = new QName(Constants.MESSAGE_TNS, "OrderNumber");
        Element orderNumberElement = DOMUtils.getChildElement(request, childName);
        String orderNumber = DOMUtils.getText(orderNumberElement);
        Document owner = documentBuilder.newDocument();

        // look-up the order. if the order is not found, we throw a
        // fault declared on interface implemented by this service

        CustomerOrder customerOrder = customerOrderDAO.selectCustomerOrder(orderNumber);
        if (customerOrder == null) {
            Element payload = owner.createElementNS(Constants.MESSAGE_TNS, "OrderNotFoundFault");
            payload.setAttribute("orderNumber", orderNumber);
            throw new OrderNotFoundFault(payload);
        }

        // we return the status on our system, i.e. warehouse status codes
        // are not relevant to customer

        Element payload = owner.createElementNS(Constants.MESSAGE_TNS, "OrderStatusResponse");
        payload.setAttribute("orderNumber", orderNumber);
        payload.setAttribute("status", customerOrder.getStatus().toString().toLowerCase());
        return payload;

    }
    
    /**
     * Selects partner application, i.e. an executable instance of WarehouseManger application,
     * based upon closest proximity to customer zipcode, places order and returns response.
     * <p>
     * Note that this method is used by the <code>OrderController</code> defined within
     * retail web application which is deployed as a part of this application (as a web
     * module), i.e. the method is not used by <code>SpringEngine</code>.
     * 
     * @param customerOrder the orderNumber
     */
    public Status getOrderStatus(String orderNumber) throws OrderNotFoundFault {

        // look-up the order. if the order is not found, we throw a
        // fault declared on interface implemented by this service

        CustomerOrder customerOrder = customerOrderDAO.selectCustomerOrder(orderNumber);
        if (customerOrder == null) {
            Document owner = documentBuilder.newDocument();
            Element payload = owner.createElementNS(Constants.MESSAGE_TNS, "OrderNotFoundFault");
            payload.setAttribute("orderNumber", orderNumber);
            throw new OrderNotFoundFault(payload);
        }

        // we return the status on our system, i.e. warehouse status codes
        // are not relevant to customer

        return customerOrder.getStatus();

    }

    /**
     * Receives shipment notification from warehouse that fulfilled the order. Completes order
     * by updating status to complete and by notifying customer via email that order has
     * shipped.
     * 
     * @param myOperation
     * @param requestContext
     */
    public void completeOrder(MyOperation myOperation, ActionContext requestContext) {

        Message requestMessage = requestContext.getMessage();
        String orderNumber = requestMessage.evaluateString(Constants.ORDER_NUMBER);
        CustomerOrder customerOrder = customerOrderDAO.selectCustomerOrder(orderNumber);

        String temp = requestMessage.getContent().getAttribute("actualShipDate");
        Date actualShipDate = null;
        try {
            actualShipDate = DateFormat.getDateInstance().parse(temp);
        } catch (ParseException pe) {
            log.error("Error completing order. " + pe.getMessage());
            customerOrder.setError(pe.getMessage());
            customerOrder.setStatus(Status.ERROR);
            return;
        }

        customerOrder.setActualShipDate(actualShipDate);
        customerOrder.setStatus(Status.COMPLETE);
        customerOrderDAO.updateCustomerOrder(customerOrder);

        // TODO: using email address on customer order. send shipment
        // notification to customer.

    }

    /**
     * Selects partner application, i.e. an executable instance of WarehouseManger application,
     * based upon closest proximity to customer zipcode, places order and returns response.
     * <p>
     * Note that this method is used by the <code>OrderController</code> defined within
     * retail web application which is deployed as a part of this application (as a web
     * module), i.e. the method is not used by <code>SpringEngine</code>.
     * 
     * @param customerOrder
     *        the 'domain' object bound by spring framework and passed by controller
     */
    public CustomerOrder placeOrder(CustomerOrder customerOrder) throws OutOfStockFault, SystemError {

        // because this method is invoked by OrderController, we use partner
        // object which models our parent application to ensure that request
        // and corresponding response gets processed by components within the
        // eoa binding layer, e.g. binding modules such as the ws-a module
        // and by components within the application layer, e.g. application
        // modules

        PartnerApplication me = myApplication.getPartner(myApplication.getName());
        PartnerRole myRole = me.getPartnerRole(ROLE_NAME);

        // retrieve object which models wsdl operation and which maps to the
        // placeOrder method defined below. populate request using values
        // bound to customer order domain object

        PartnerOperation op = myRole.createOperation(Constants.PLACE_ORDER_OPERATION);
        ActionContext requestContext = op.createRequestAction();
        Message requestMessage = requestContext.getMessage();
        String zip = customerOrder.getZipCode();
        DOMUtils.setText((Element)requestMessage.evaluateNode(Constants.ZIP_CODE), zip);
        Element itemsEl = (Element)requestMessage.evaluateNode(Constants.ITEMS);
        for (Item item : customerOrder.getItems()) {
            Element itemEl = requestMessage.createElementNS(Constants.MESSAGE_TNS, "Item");
            itemEl.setAttribute("id", item.getId());
            itemEl.setAttribute("quantity", Integer.toString(item.getQuantity()));
            itemsEl.appendChild(itemEl);
        }

        // invoke operation which ultimately ends up invoking method below.
        // if result is a fault context instance, then it either a system
        // fault, or the only declared fault - an out of stock fault
        
        ActionContext responseContext = op.sendRequest(requestContext, 0);
        Message responseMessage = responseContext.getMessage();
        if (responseContext instanceof FaultContext) {
            ApplicationFault applicationFault = ((FaultContext)responseContext).getFault();
            if (applicationFault.getFaultName().equals(SystemFault.FAULT_NAME)) {
                String reason = responseMessage.evaluateString(SystemFault.FAULT_REASON_PROPERTY);
                throw new SystemError(reason);
            } else {
                Element payload = (Element)responseMessage.evaluateNode(Constants.ITEMS);
                throw new OutOfStockFault(payload);
            }
        }
        
        // the order was placed. extract parameters from message, update
        // our customer order object and return to controller
        
        String orderNumber = responseMessage.evaluateString(Constants.ORDER_NUMBER);
        customerOrder.setOrderNumber(orderNumber);
        Element payload = responseMessage.getContent();
        Double shippingCost = null;
        try {
            String text = payload.getAttribute("shippingCost");
            shippingCost = (Double)NumberFormat.getInstance().parse(text);
        } catch (ParseException pe) {
            log.error("Error completing order. " + pe.getMessage());
            throw new RuntimeException(pe.getMessage());
        }
        customerOrder.setShipCost(shippingCost);
        Date estimatedShipDate = null;
        try {
            String temp = payload.getAttribute("estimatedShipDate");
            estimatedShipDate = DateFormat.getDateInstance().parse(temp);
            customerOrder.setEstShipDate(estimatedShipDate);
        } catch (ParseException pe) {
            log.error("Error completing order. " + pe.getMessage());
            throw new RuntimeException(pe.getMessage());
        }
        
        return customerOrder;
        
    }

    /*
     * dependency injector. set by spring framework
     */
    public void setCustomerOrderDAO(CustomerOrderDAO customerOrderDAO) {
        this.customerOrderDAO = customerOrderDAO;
    }

    /*
     * dependency injector. set by spring framework
     */
    public void setCatalogProcessor(CatalogProcessor catalogProcessor) {
        this.catalogProcessor = catalogProcessor;
    }

    /*
     * Life cycle method invoked by spring framework
     */
    public void init() throws Exception {

        // set up document builder factory which is used within method
        // that demonstrates a blocking invocation style

        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        dbf.setNamespaceAware(true);
        documentBuilder = dbf.newDocumentBuilder();

        myApplication = MyApplicationContext.getApplication();
        myRole = myApplication.getMyRole(ROLE_NAME);

    }

    /*
     * Life cycle method invoked by spring framework
     */
    public void destroy() {
    }

}
